//----------------------------------------------------------------------
//
// NTSID
//
// Copyright (c) 1997 Mark Russinovich and Bryce Cogswell
//
// Changes the computer SID. 
//
// This code is protected under copyright law. You do not have 
// permission to use this code in a commercial SID-changing product.
//
//----------------------------------------------------------------------
#define UNICODE 1
#include <windows.h>
#include <commctrl.h>
#include <stdio.h>
#include <tchar.h>
#include <process.h>
#include "resource.h"  

//
// Defines for security buffer sizes
//
#define SZ_REL_SD_BUF		8192
#define SZ_ABS_SD_BUF		8192

//
// Max length of path name displayed in
// progress dialogs
//
#define MAX_DLG_PATHLEN		55

//
// ProductType name - this isn't exact, but its
// good enough
//
#define PRODUCTTYPEPATH		L"System\\C"
#define PRODUCTTYPENAME		L"ProductOptions"

//
// Profile list path
//
#define PROFILELISTPATH L"Software\\Microsoft\\Windows NT\\CurrentVersion\\ProfileList"

//
// General buffer for storing temporary strings
//
static TCHAR		msgbuf[ 1024 ];
   
//
// The binary representations of the old and the
// new subauthority fields of the computer SID
//
CHAR					NewSubAuth[16];
CHAR					OldSubAuth[16];
DWORD					SubAuthLength;


//
// Application instance handle
//
HINSTANCE			hInst;

//
// Values the main dialog can return, which indicate
// whether to proceed or not.
//
enum {
	DIALOG_OK,
	DIALOG_CANCEL,
};


//
// Forwards
//
LONG APIENTRY MainDialog( HWND hDlg, UINT message, UINT wParam, LONG lParam );


//===================================================================
//
// SECONDARY DIALOG PROCEDURES
//
//===================================================================

//----------------------------------------------------------------------
//  
// AboutDlgProc
//
// Pops up the standard About dialog box.
//
//----------------------------------------------------------------------
BOOL CALLBACK AboutDlgProc (HWND hDlg, UINT message, UINT wParam,
                       		LONG lParam) {
	switch (message) {
	case WM_INITDIALOG :
		return TRUE ;
 
	case WM_COMMAND :
		switch (wParam) {

		case IDOK :
			EndDialog (hDlg, 0) ;
			return TRUE ;
		}
		break; 

	case WM_CLOSE:	
		EndDialog (hDlg, 0);
		return TRUE;
	}
	return FALSE ;
}

//----------------------------------------------------------------------
//
// UpdateProgressDialog
//
// Make the pathname fit into the dialog box
//
//----------------------------------------------------------------------
void UpdateProgressDialog( HWND hDlg, DWORD Control, PTCHAR PathName )
{
	TCHAR		shortName[MAX_DLG_PATHLEN];
	DWORD		i, j, k;

	if( (i = wcslen( PathName )) > MAX_DLG_PATHLEN ) {

		j = MAX_DLG_PATHLEN-2;
		shortName[MAX_DLG_PATHLEN-1] = 0;
		while( j > (MAX_DLG_PATHLEN/2+3) ) {

			shortName[j] = PathName[i];
			j--;
			i--;
		}

		for( k = 0; k < MAX_DLG_PATHLEN/2; k++ ) {

		   shortName[k] = PathName[k];
		}

		while( k <= j ) shortName[k++] = L'.';


		SetDlgItemText( hDlg, Control, shortName );

	} else {

		SetDlgItemText( hDlg, Control, PathName );
	}
}


//----------------------------------------------------------------------
//  
// UpdateDlgProc
//
// Dialog box shown while file systems or the Registry is
//  being scanned for SID update.
//
//----------------------------------------------------------------------
BOOL CALLBACK UpdateDlgProc (HWND hDlg, UINT message, UINT wParam,
                       		LONG lParam) {
	switch (message) {
	case WM_INITDIALOG :

		return TRUE ;
 	}
	return FALSE ;
}

//===================================================================
//
// RANDOM NUMBER GENERATION FOR MAKING A NEW SID
//
//===================================================================

//----------------------------------------------------------------------
//
// Random
//
// 32-bit pseudo-random numbers
//
//----------------------------------------------------------------------
volatile BOOL RandomExpired;

void __cdecl RandomTrigger( void * arg )
{
	// Wait until main thread is ready to go
	while ( RandomExpired )
		Sleep( 1 );
	// Now sleep while main thread counts
	Sleep( 1 );
	// Notify main thread that time is up
	RandomExpired = TRUE;
	// Exit
	_endthread();
}

DWORD RandomBits( void )
{
	register DWORD cnt = 0;

	// Tell trigger we're not ready yet
	RandomExpired = TRUE;
	// Start trigger
	_beginthread( RandomTrigger, 0, NULL );
	// Okay, tell it to start sleeping
	RandomExpired = FALSE;
	// Count as fast as we can
	while ( !RandomExpired )  {
		++cnt;
	}
	// Only low bits are accurate!
	return cnt;
}


DWORD Random( void )
{
	return	RandomBits() << 28	^
			RandomBits() << 24	^
			RandomBits() << 20	^
			RandomBits() << 16	^
			RandomBits() << 12	^
			RandomBits() << 8	^
			RandomBits() << 4	^
			RandomBits() << 0;
}


//===================================================================
//
// REGISTRY SECURITY CONVENIENCE ROUTNES 
//
//===================================================================


//----------------------------------------------------------------------
//
// GetRegAccess
//
// Returns the original security descriptor of a key.
//
//----------------------------------------------------------------------
BOOLEAN GetRegAccess( HKEY hKey, PSECURITY_DESCRIPTOR *SecDesc  )
{
	LONG		status;
	DWORD		nb;

	//
	// Get access
	//
	nb = 0;
	status = RegGetKeySecurity( hKey, DACL_SECURITY_INFORMATION, *SecDesc, &nb );
	*SecDesc = malloc( nb );
	status = RegGetKeySecurity( hKey, DACL_SECURITY_INFORMATION, *SecDesc, &nb );
	if ( status != ERROR_SUCCESS )
		return FALSE;
	return TRUE;
}


//----------------------------------------------------------------------
//
// SetRegAccess
//
// Sets the key with the specified security descriptor. If the key's
// name is passed (Rooot != NULL), then the key is re-opened to give
// the caller updated access rights that reflect the changed security.
//
//----------------------------------------------------------------------
BOOLEAN SetRegAccess( HKEY Root, TCHAR *path,
					 PSECURITY_DESCRIPTOR SecDesc, HKEY *hKey )
{
	LONG		status;

	//
	// Grant requested access
	//
	status = RegSetKeySecurity( *hKey, DACL_SECURITY_INFORMATION, SecDesc );
	if ( status != ERROR_SUCCESS )
		return FALSE;

	//
	// Re-open the key if requested
	//
	if( Root ) {

		RegCloseKey( *hKey );
		status = RegOpenKey( Root, path, hKey );
		if( status != ERROR_SUCCESS ) 
			return FALSE;
	}
	return TRUE;
}


//----------------------------------------------------------------------
//
// GetRegSecDesc
//
// Gets the security descriptor from the specified key.
//
//----------------------------------------------------------------------
PSECURITY_DESCRIPTOR GetRegSecDesc( HKEY Root, TCHAR * path, 
								SECURITY_INFORMATION Information ) 
{
	HKEY					hKey;
	LONG					status;
	DWORD					nb;
	PSECURITY_DESCRIPTOR	SecDesc;

	//
	// Open the key with no access requests, since we don't need
	// any.
	//
	status = RegOpenKeyEx( Root, path, 0, KEY_READ, &hKey );
	if ( status != ERROR_SUCCESS )
		return NULL;

	//
	// Grab a copy of the security for key
	//
	nb = 0;
	SecDesc = NULL;
	status = RegGetKeySecurity( hKey, Information, SecDesc, &nb );
	SecDesc = malloc( nb );
	status = RegGetKeySecurity( hKey, Information, SecDesc, &nb );
	if ( status != ERROR_SUCCESS )
		return NULL;

	RegCloseKey( hKey );
	return SecDesc;
}


//----------------------------------------------------------------------
//
// SecurityReplaceSID
//
// This convenience function scans the passed in SID for an instance
// of the old computer security SID, and if it finds it, it updates
// the instance with the new computer SID.
//
//----------------------------------------------------------------------
BOOLEAN SecurityReplaceSID( PSID pSid )
{
	BOOLEAN		foundSid;
	PUCHAR		numSubAuths;
	PDWORD		firstSubAuth;
	DWORD		i;

	foundSid = FALSE;
	numSubAuths = GetSidSubAuthorityCount(pSid);
	if( (DWORD) *numSubAuths >= SubAuthLength/sizeof(DWORD) ) {

		firstSubAuth = GetSidSubAuthority(pSid, 0); 
		for( i=0; i < (DWORD) *numSubAuths - SubAuthLength/sizeof(DWORD) + 1; i++ ) {

			if( !memcmp( &firstSubAuth[i], OldSubAuth, SubAuthLength )) {

				//
				// Found it - change it and continue
				//
				memcpy( &firstSubAuth[i], NewSubAuth, SubAuthLength );
				foundSid = TRUE;
				break;
			}
		}
	}
	return foundSid;
}

//===================================================================
//
// REGISTRY SECURITY DESCRIPTOR UPDATING 
//
//===================================================================

//----------------------------------------------------------------------
//
// CheckKeyOwnershipForSID
//
// Sees if the file ownership SID contains the old computer SID, and
// if so, replaces it.
//
//----------------------------------------------------------------------
void CheckKeyOwnershipForSID(HKEY hKey, PSECURITY_DESCRIPTOR psdRegSD )
{
	PSID		psidRegOwnerSID;
	PSID		psidRegGroupSID;
	BOOL		bOwnerDefaulted;

	//
	// Get the owner SID
	//
	if (!GetSecurityDescriptorOwner(psdRegSD,
		(PSID *)&psidRegOwnerSID, (LPBOOL)&bOwnerDefaulted) ||
		!psidRegOwnerSID) { 

		return;
	}

	//
	// If the old SID is in the owner, we have to write
	// the updated owner to the key
	//
    if( SecurityReplaceSID( psidRegOwnerSID ) ) {
		if (!RegSetKeySecurity(hKey,
				(SECURITY_INFORMATION)OWNER_SECURITY_INFORMATION,
				psdRegSD)){
			
			return;
		}
    }

	//
	// Now get the Group SID and do the same thing
	//
	if (!GetSecurityDescriptorGroup(psdRegSD,
		(PSID *)&psidRegGroupSID, (LPBOOL)&bOwnerDefaulted) ||
		!psidRegOwnerSID) { 

		return;
	}

	//
	// If the old SID is in the owner, we have to write
	// the updated owner to the key
	//
    if( SecurityReplaceSID( psidRegGroupSID ) ) {
		if (!RegSetKeySecurity(hKey,
				(SECURITY_INFORMATION)GROUP_SECURITY_INFORMATION,
				psdRegSD)){
			
			return;
		}
    }
	return;
}


//----------------------------------------------------------------------
//
// CheckKeyACLForSID
//
// Scan's the security descriptor's ACEs, looking for instances
// of the old computer SID.
//
//----------------------------------------------------------------------
void CheckKeyACLForSID( HKEY hKey, BOOLEAN Dacl, 
					   PSECURITY_DESCRIPTOR psdRegSD )
{
	PACL       paclReg;
	BOOL       bHasACL;
	BOOL       bOwnerDefaulted;
	DWORD      dwAcl_i;
	BOOLEAN    descriptorModified;
	DWORD      dwLastError   = NO_ERROR;
	ACL_SIZE_INFORMATION asiAclSize;
	DWORD      dwBufLength = sizeof(asiAclSize);
	ACCESS_ALLOWED_ACE   *paaAllowedAce;

	if( Dacl ) {
		if (!GetSecurityDescriptorDacl(psdRegSD,
							 (LPBOOL)&bHasACL, (PACL *)&paclReg,
							 (LPBOOL)&bOwnerDefaulted)) {
			return;
		}
	} else {
		if (!GetSecurityDescriptorSacl(psdRegSD,
							 (LPBOOL)&bHasACL, (PACL *)&paclReg,
							 (LPBOOL)&bOwnerDefaulted)) {
			return;
		}
	}

	//
	// If no ACL to process, so OK, return
	//
	if (!bHasACL || !paclReg ) return;

	if (!GetAclInformation(paclReg, (LPVOID)&asiAclSize,
				 (DWORD)dwBufLength, (ACL_INFORMATION_CLASS)AclSizeInformation)){
		return;
	}

	// 
	// Look through the ACEs
	//
	descriptorModified = FALSE;
	for (dwAcl_i = asiAclSize.AceCount-1;  ((int)dwAcl_i) >= 0;  dwAcl_i--) {

		//
		// If we can't get an ACE, bail
		//
		if (!GetAce(paclReg, dwAcl_i, (LPVOID *)&paaAllowedAce)) {

			return;
		}

		//
		// Make sure we're dealing with an ACE we know
		//
		if (!( (paaAllowedAce->Header.AceType == ACCESS_ALLOWED_ACE_TYPE)
				||(paaAllowedAce->Header.AceType == ACCESS_DENIED_ACE_TYPE )
				||(paaAllowedAce->Header.AceType == SYSTEM_AUDIT_ACE_TYPE  )
				||(paaAllowedAce->Header.AceType == SYSTEM_ALARM_ACE_TYPE  ))) {
			continue;
		}

		//
		// Look at the SID's subauthorities to see if there's a match
		// with the old computer SID
		//
		descriptorModified |= SecurityReplaceSID((PSID)&(paaAllowedAce->SidStart)); 
	}

	//
	// If the security descriptor was modified because an
	// old computer SID was converted to the new one, write
	// the new descriptor to disk.
	//
	if( descriptorModified ) {

		//
		// Modify the SD on the hard disk
		//
		if( Dacl ) {

			//
			// Modify the SD in virtual memory
			//
			if (!SetSecurityDescriptorDacl(psdRegSD,
								TRUE, paclReg, FALSE)) {

				return;
			}

			if (!RegSetKeySecurity(hKey,
							(SECURITY_INFORMATION)DACL_SECURITY_INFORMATION,
							psdRegSD)){

				return;
			}
		} else {

			//
			// Modify the SD in virtual memory
			//
			if (!SetSecurityDescriptorSacl(psdRegSD,
								TRUE, paclReg, FALSE)) {

				return;
			}

			if (!RegSetKeySecurity(hKey,
							(SECURITY_INFORMATION)SACL_SECURITY_INFORMATION,
							psdRegSD)){

				return;
			}
		}
	}
}


//----------------------------------------------------------------------
//
// CheckKeySID
//
// Reads the key's security descriptor and converts it to absolute
// format. Then it calls a function to check the ownership SIDs and
// finally one to check the DACL SIDs, to see if they need to be
// updated.
//
//----------------------------------------------------------------------
BOOL CheckKeySID( HKEY hKey, HKEY Root, LPTSTR lpszFullName,
				 PSECURITY_DESCRIPTOR psdSrelRegSD )
{
	UCHAR       ucBufAbs[SZ_ABS_SD_BUF];
	UCHAR       ucBufCtrl[sizeof(PSECURITY_DESCRIPTOR_CONTROL)];
	DWORD       dwSDLength = SZ_REL_SD_BUF;
	PSECURITY_DESCRIPTOR psdAbsRegSD = (PSECURITY_DESCRIPTOR)&ucBufAbs;
	PSECURITY_DESCRIPTOR_CONTROL psdcCtrl = (PSECURITY_DESCRIPTOR_CONTROL)&ucBufCtrl;
	PACL        paclDacl;
	PACL        paclSacl;
	PSID        psidSidOwn;
	PSID        psidSidPG;
	BOOL        bDaclPresent;
	BOOL        bDaclDefaulted;
	BOOL        bSaclPresent;
	BOOL        bSaclDefaulted;
	BOOL        bOwnerDefaulted;
	BOOL        bGroupDefaulted;
	BOOL        bSDSelfRelative;
	DWORD       dwRevision;

	//
	// Build File SD in absolute format for potential later modification
	//
	if (!InitializeSecurityDescriptor(psdAbsRegSD,
		 SECURITY_DESCRIPTOR_REVISION)) { 
		return FALSE;
	}

	//
	// Get Control from relative format File SD
	//
	if (!GetSecurityDescriptorControl(psdSrelRegSD,
			psdcCtrl,
			&dwRevision)) { 
		return FALSE;
	} 
	bSDSelfRelative = (SE_SELF_RELATIVE & *psdcCtrl);

	//
	// Set DACL into absolute format File SD
	//
	if (bDaclPresent = (SE_DACL_PRESENT   & *psdcCtrl)) {

		bDaclDefaulted = (SE_DACL_DEFAULTED & *psdcCtrl);
	}

	if (!GetSecurityDescriptorDacl(psdSrelRegSD,
							&bDaclPresent,      // fDaclPresent flag
							&paclDacl,
							&bDaclDefaulted)) {

		return FALSE;
	}
	if (!SetSecurityDescriptorDacl(psdAbsRegSD,
							bDaclPresent,       // fDaclPresent flag
							paclDacl,
							bDaclDefaulted)) {

		return FALSE;
	}

	//  
	// Set SACL into absolute format File SD
	//
	if (bSaclPresent = (SE_SACL_PRESENT   & *psdcCtrl)){

		bSaclDefaulted = (SE_SACL_DEFAULTED & *psdcCtrl);
	}

	if (!GetSecurityDescriptorSacl(psdSrelRegSD,
						&bSaclPresent,      // fSaclPresent flag
						&paclSacl,
						&bSaclDefaulted)) {

		return FALSE;
	}
	if (!SetSecurityDescriptorSacl(psdAbsRegSD,
						bSaclPresent,       // fSaclPresent flag
						paclSacl,
						bSaclDefaulted)) {

		return FALSE;
	} 

	//
	// Set Owner into absolute format File SD
	//
	bOwnerDefaulted = (SE_OWNER_DEFAULTED & *psdcCtrl);
	if (!GetSecurityDescriptorOwner(psdSrelRegSD,
					&psidSidOwn,
					&bOwnerDefaulted)) {

		return FALSE;
	}
	if (!SetSecurityDescriptorOwner(psdAbsRegSD,
					psidSidOwn,
					bOwnerDefaulted)) {

		return FALSE;
	}

	//
	// Set Group into absolute format File SD
	//
	bGroupDefaulted = (SE_GROUP_DEFAULTED & *psdcCtrl);
	if (!GetSecurityDescriptorGroup(psdSrelRegSD,
					&psidSidPG,
					&bGroupDefaulted)) {

		return FALSE;
	}
	if (!SetSecurityDescriptorGroup(psdAbsRegSD,
					psidSidPG,
					bGroupDefaulted)) {

		return FALSE;
	}

	//
	// Now we can see if the old computer SID is embedded
	// in the security descriptor, and change it if necessary
	//
	CheckKeyOwnershipForSID(hKey, psdAbsRegSD );
	
	//
	// Check to see if SID is embedded in the DACL and then
	// SACL
	//
	CheckKeyACLForSID( hKey, TRUE, psdAbsRegSD );
	CheckKeyACLForSID( hKey, FALSE, psdAbsRegSD );

	//
	// Always reset the DACL and close the key
	//
	SetRegAccess( NULL, NULL, psdAbsRegSD, &hKey );
	RegCloseKey( hKey );
	free( psdSrelRegSD );

	return(TRUE);
}


//----------------------------------------------------------------------
//
// UpdateKeySID
//
// Updates the SIDs for the specified registry key and recurses
// into subkeys.
//
//----------------------------------------------------------------------
void UpdateKeySID( HWND hDlg, HKEY Root, PTCHAR RootName, PTCHAR PathName,
				  PSECURITY_DESCRIPTOR newSecDesc )
{
	TCHAR			subKeyName[ 1024 ];
	TCHAR			subRootName[ 1024 ];
	TCHAR			subPathName[ 1024 ];
	HKEY			hKey;
	FILETIME		Time;
	DWORD			subKeys;
	int				idx;
	DWORD			status, nb;
	PSECURITY_DESCRIPTOR	origDesc;

	//
	// Update the progress dialog
	//
	UpdateProgressDialog( hDlg, IDC_PATHNAME, RootName );

	//
	// Change the security on the key so that we can enumerate it
	//
	status = RegOpenKeyEx( Root, PathName, 0, 
						WRITE_DAC|READ_CONTROL|
						ACCESS_SYSTEM_SECURITY,
						&hKey );
	if( status != ERROR_SUCCESS ) {
		return;
	}

	//
	// Get the source security and set the new descriptor
	//
	GetRegAccess( hKey, &origDesc );
	SetRegAccess( Root, PathName, newSecDesc, &hKey );

	//
	// See if there are subkeys to enumerate
	//
	status = RegQueryInfoKey( hKey, NULL, NULL, NULL,
							&subKeys, NULL,
							NULL, NULL, NULL, NULL,
							NULL, NULL );
	if( status == ERROR_SUCCESS && subKeys ) {
		//
		// Allocate buffers 
		//

		idx = 0;
		while(1)  {

			nb = sizeof subKeyName;
			status = RegEnumKeyEx( hKey, idx, subKeyName, &nb, NULL, NULL, NULL, &Time );
			if ( status == ERROR_NO_MORE_ITEMS )
				break;
			if ( status != ERROR_SUCCESS )
				break;


			//
			// Construct the names
			//
			wsprintf( subRootName, L"%s\\%s", RootName, subKeyName );
			if( wcscmp( PathName, L"" )) 
				wsprintf( subPathName, L"%s\\%s", PathName, subKeyName );
			else
				wcscpy( subPathName, subKeyName );

			//
			// Process the subkey - leave the producttype alone, since
			// we would otherwise getting an annoying popup about
			// messing with our license.
			//
			if( !( !wcsnicmp( subPathName, 
						PRODUCTTYPEPATH,
						wcslen(PRODUCTTYPEPATH)) &&
					!wcsicmp( subKeyName, 
						PRODUCTTYPENAME))) {
				UpdateKeySID( hDlg, Root,  subRootName, subPathName,
							  newSecDesc );
			}
			idx++;
		}
	}

	//
	// Now process the security on this key. This
	// will restore the security either to updated
	// (SID converted) settings, or the original
	// settings of the key. It will also close
	// the key.
	//
	CheckKeySID( hKey, Root, PathName, origDesc );
}

//----------------------------------------------------------------------
//
// UnloadProfileHives
//
// Unloads the user hives that were loaded into the Registry.
//
//----------------------------------------------------------------------
void UnloadProfileHives( DWORD NumLoadedHives )
{
	TCHAR		hiveName[256];
	DWORD		i;

	for( i = 0; i < NumLoadedHives; i++ ) {

		wsprintf( hiveName, L"User%d", i );
		RegUnLoadKey( HKEY_LOCAL_MACHINE, hiveName );
	}
}

//----------------------------------------------------------------------
//
// LoadProfileHives
//
// Loads all the user profile hives into the Registry, so that
// their security descriptors can be updated.
//
//----------------------------------------------------------------------
void LoadProfileHives( DWORD *NumLoadedHives )
{
	int				idx;
	DWORD			status, nb, type;
	HKEY			hKey, hProfileKey;
	TCHAR			profileKeyPath[1024], hiveName[256];
	TCHAR			subKeyName[1024], imagePath[1024], hivePath[1024];
	TCHAR			systemRoot[256];

	//
	// Get the system root path
	//
	status = RegOpenKey( HKEY_LOCAL_MACHINE,
			L"Software\\Microsoft\\Windows NT\\CurrentVersion", &hKey );
	if( status == ERROR_SUCCESS ) {

		nb = 1024;
		(void) RegQueryValueEx( hKey, L"SystemRoot",
								NULL, &type, (PCHAR) systemRoot, &nb );
		RegCloseKey( hKey );
	}

	//
	// First, locate the user hives and load them
	//
	*NumLoadedHives = 0;
	status = RegOpenKey( HKEY_LOCAL_MACHINE, PROFILELISTPATH, &hKey );
	if( status == ERROR_SUCCESS ) {

		idx = 0;
		while(1)  {

			nb = 1024;
			status = RegEnumKeyEx( hKey, idx, subKeyName, &nb, NULL, NULL, NULL, NULL );
			if ( status == ERROR_NO_MORE_ITEMS )
				break;
			if ( status != ERROR_SUCCESS )
				break;

			//
			// Open the subkey and get the location of the registry hive
			//
			wsprintf( profileKeyPath, L"%s\\%s", PROFILELISTPATH, subKeyName );
			status = RegOpenKey( HKEY_LOCAL_MACHINE, profileKeyPath, &hProfileKey );
			if( status == ERROR_SUCCESS ) {
				
				nb = 1024;
				status = RegQueryValueEx( hProfileKey, L"ProfileImagePath",
										NULL, &type, (PCHAR) imagePath, &nb );
				if( status == ERROR_SUCCESS ) {

					//
					// Expand the system root
					//
					if( !wcsnicmp( imagePath, L"%SystemRoot%", 
										wcslen(L"%SystemRoot%"))) {
					
						wsprintf( hivePath, L"%s\\%s\\ntuser.dat",
										systemRoot,
										imagePath + wcslen(L"%SystemRoot%"));
					} else {

						wsprintf( hivePath, L"%s\\ntuser.dat",
										imagePath );
					}

					//
					// Now load the hive
					//
					wsprintf( hiveName, L"User%d", *NumLoadedHives );
					status = RegLoadKey( HKEY_LOCAL_MACHINE, hiveName, hivePath );
					if( status == ERROR_SUCCESS ) {

						(*NumLoadedHives)++;
					}	
				}
			}
			idx++;
		}
		RegCloseKey( hKey );
	}
}


//----------------------------------------------------------------------
//
// UpdateRegistrySID
//
// Scans all the Registry hives looking for SIDs
// that need to be changed in the security descriptors.
//
//----------------------------------------------------------------------
void UpdateRegistrySID( PSECURITY_DESCRIPTOR newSecDesc )
{
	TCHAR			startPath[16];
	HWND			regDialog;
	DWORD			NumLoadedHives;

	//
	// First, load all the user profile hives
	//
	LoadProfileHives( &NumLoadedHives );
	
	//
	// Create a dialog to show the file traversal
	//
	regDialog = CreateDialog( hInst, L"REGISTRY", NULL,
							(DLGPROC) UpdateDlgProc );

	//
	// Update all the loaded root keys
	//
	wcscpy( startPath, L"");
	UpdateKeySID( regDialog, HKEY_LOCAL_MACHINE, L"HKLM", startPath, newSecDesc ); 	
	wcscpy( startPath, L"");
	UpdateKeySID( regDialog, HKEY_CURRENT_USER, L"HKCU", startPath, newSecDesc );

	//
	// Delete the file dialog
	//
	DestroyWindow( regDialog );

	//
	// Unload user profile hives
	//
	UnloadProfileHives( NumLoadedHives );
}


//===================================================================
//
// FILE SECURITY DESCRIPTOR UPDATING 
//
//===================================================================


//----------------------------------------------------------------------
//
// CheckFileOwnershipForSID
//
// Sees if the file ownership SID contains the old computer SID, and
// if so, replaces it.
//
//----------------------------------------------------------------------
void CheckFileOwnershipForSID(PSECURITY_DESCRIPTOR psdFileSD, 
							  LPTSTR  lpszFullName)
{
	PSID		psidFileOwnerSID;
	PSID		psidFileGroupSID;
	BOOL		bOwnerDefaulted;

	//
	// Get the owner SID
	//
	if (!GetSecurityDescriptorOwner(psdFileSD,
		(PSID *)&psidFileOwnerSID, (LPBOOL)&bOwnerDefaulted)) { 

		return;
	}

	//
	// If the old SID is in the owner, we have to write
	// the updated owner to disk
	//
    if( SecurityReplaceSID( psidFileOwnerSID) ) {
		if (!SetFileSecurity(lpszFullName,
            (SECURITY_INFORMATION)OWNER_SECURITY_INFORMATION,
            psdFileSD)){
			
			return;
		}
    }

	//
	// Now do the same with the Group SID
	//
	if (!GetSecurityDescriptorOwner(psdFileSD,
		(PSID *)&psidFileGroupSID, (LPBOOL)&bOwnerDefaulted)) { 

		return;
	}

	//
	// If the old SID is in the owner, we have to write
	// the updated owner to disk
	//
    if( SecurityReplaceSID( psidFileGroupSID) ) {
		if (!SetFileSecurity(lpszFullName,
            (SECURITY_INFORMATION)GROUP_SECURITY_INFORMATION,
            psdFileSD)){
			
			return;
		}
    }
	return;
}


//----------------------------------------------------------------------
//
// CheckACLForSID
//
// Scan's the security descriptor's ACEs, looking for instances
// of the old computer SID.
//
//----------------------------------------------------------------------
void CheckFileACLForSID( BOOLEAN Dacl, PSECURITY_DESCRIPTOR psdFileSD,
                             LPTSTR  lpszFullName)
{
	PACL       paclFile;
	BOOL       bHasACL;
	BOOL       bOwnerDefaulted;
	DWORD      dwAcl_i;
	BOOLEAN    descriptorModified;
	DWORD      dwLastError   = NO_ERROR;
	ACL_SIZE_INFORMATION asiAclSize;
	DWORD      dwBufLength = sizeof(asiAclSize);
	ACCESS_ALLOWED_ACE   *paaAllowedAce;

	if( Dacl ) {
		if (!GetSecurityDescriptorDacl(psdFileSD,
							 (LPBOOL)&bHasACL, (PACL *)&paclFile,
							 (LPBOOL)&bOwnerDefaulted)) {
			return;
		}
	} else {
		if (!GetSecurityDescriptorSacl(psdFileSD,
							 (LPBOOL)&bHasACL, (PACL *)&paclFile,
							 (LPBOOL)&bOwnerDefaulted)) {
			return;
		}
	}

	//
	// If no ACL to process, so OK, return
	//
	if (!bHasACL || !paclFile ) return;

	if (!GetAclInformation(paclFile, (LPVOID)&asiAclSize,
				 (DWORD)dwBufLength, (ACL_INFORMATION_CLASS)AclSizeInformation)){
		return;
	}
	descriptorModified = FALSE;

	// 
	// Look through the ACEs
	//
	for (dwAcl_i = asiAclSize.AceCount-1;  ((int)dwAcl_i) >= 0;  dwAcl_i--) {

		//
		// If we can't get an ACE, bail
		//
		if (!GetAce(paclFile, dwAcl_i, (LPVOID *)&paaAllowedAce)) {

			return;
		}

		//
		// Make sure we're dealing with an ACE we know
		//
		if (!( (paaAllowedAce->Header.AceType == ACCESS_ALLOWED_ACE_TYPE)
				||(paaAllowedAce->Header.AceType == ACCESS_DENIED_ACE_TYPE )
				||(paaAllowedAce->Header.AceType == SYSTEM_AUDIT_ACE_TYPE  )
				||(paaAllowedAce->Header.AceType == SYSTEM_ALARM_ACE_TYPE  ))) {
			continue;
		}

		//
		// Look at the SID's subauthorities to see if there's a match
		// with the old computer SID
		//
		descriptorModified |= SecurityReplaceSID((PSID)&(paaAllowedAce->SidStart));
	}

	//
	// If the security descriptor was modified because an
	// old computer SID was converted to the new one, write
	// the new descriptor to disk.
	//
	if( descriptorModified ) {

		//
		// Modify the SD on the hard disk
		//
		if( Dacl ) {

			//
			// Modify the SD in virtual memory
			//
			if (!SetSecurityDescriptorDacl(psdFileSD,
								TRUE, paclFile, FALSE)) {

				return;
			}

			if (!SetFileSecurity(lpszFullName,
							(SECURITY_INFORMATION)DACL_SECURITY_INFORMATION,
							psdFileSD)){

				return;
			}
		} else {

			//
			// Modify the SD in virtual memory
			//
			if (!SetSecurityDescriptorSacl(psdFileSD,
								TRUE, paclFile, FALSE)) {

				return;
			}

			if (!SetFileSecurity(lpszFullName,
							(SECURITY_INFORMATION)SACL_SECURITY_INFORMATION,
							psdFileSD)){

				return;
			}
		}
	}
}


//----------------------------------------------------------------------
//
// CheckFileSID
//
// Reads the file's security descriptor and converts it to absolute
// format. Then it calls a function to check the ownership SIDs and
// finally one to check the DACL SIDs, to see if they need to be
// updated.
//
//----------------------------------------------------------------------
BOOL CheckFileSID(LPTSTR lpszFullName)
{
	UCHAR		ucBuf[SZ_REL_SD_BUF];
	UCHAR       ucBufAbs[SZ_ABS_SD_BUF];
	UCHAR       ucBufCtrl[sizeof(PSECURITY_DESCRIPTOR_CONTROL)];
	DWORD       dwSDLength = SZ_REL_SD_BUF;
	DWORD       dwSDLengthNeeded;
	PSECURITY_DESCRIPTOR psdSrelFileSD = (PSECURITY_DESCRIPTOR)&ucBuf;
	PSECURITY_DESCRIPTOR psdAbsFileSD = (PSECURITY_DESCRIPTOR)&ucBufAbs;
	PSECURITY_DESCRIPTOR_CONTROL psdcCtrl = (PSECURITY_DESCRIPTOR_CONTROL)&ucBufCtrl;
	PACL        paclDacl;
	PACL        paclSacl;
	PSID        psidSidOwn;
	PSID        psidSidPG;
	BOOL        bDaclPresent;
	BOOL        bDaclDefaulted;
	BOOL        bSaclPresent;
	BOOL        bSaclDefaulted;
	BOOL        bOwnerDefaulted;
	BOOL        bGroupDefaulted;
	BOOL        bSDSelfRelative;
	DWORD       dwRevision;

	//
	// Get the file's security descriptor
	//
	if (!GetFileSecurity(lpszFullName,
						(SECURITY_INFORMATION)( OWNER_SECURITY_INFORMATION
					   | GROUP_SECURITY_INFORMATION
					   | DACL_SECURITY_INFORMATION
					   | SACL_SECURITY_INFORMATION),
						psdSrelFileSD,
						dwSDLength,
						(LPDWORD)&dwSDLengthNeeded)) {

		return(FALSE);
	}

	//
	// Build File SD in absolute format for potential later modification
	//
	if (!InitializeSecurityDescriptor(psdAbsFileSD,
		 SECURITY_DESCRIPTOR_REVISION)) { 
		return FALSE;
	}

	//
	// Get Control from relative format File SD
	//
	if (!GetSecurityDescriptorControl(psdSrelFileSD,
			psdcCtrl,
			&dwRevision)) { 
		return FALSE;
	} 
	bSDSelfRelative = (SE_SELF_RELATIVE & *psdcCtrl);

	//
	// Set DACL into absolute format File SD
	//
	if (bDaclPresent = (SE_DACL_PRESENT   & *psdcCtrl)) {

		bDaclDefaulted = (SE_DACL_DEFAULTED & *psdcCtrl);
	}

	if (!GetSecurityDescriptorDacl(psdSrelFileSD,
							&bDaclPresent,      // fDaclPresent flag
							&paclDacl,
							&bDaclDefaulted)) {

		return FALSE;
	}
	if (!SetSecurityDescriptorDacl(psdAbsFileSD,
							bDaclPresent,       // fDaclPresent flag
							paclDacl,
							bDaclDefaulted)) {

		return FALSE;
	}

	//  
	// Set SACL into absolute format File SD
	//
	if (bSaclPresent = (SE_SACL_PRESENT   & *psdcCtrl)){

		bSaclDefaulted = (SE_SACL_DEFAULTED & *psdcCtrl);
	}

	if (!GetSecurityDescriptorSacl(psdSrelFileSD,
						&bSaclPresent,      // fSaclPresent flag
						&paclSacl,
						&bSaclDefaulted)) {

		return FALSE;
	}
	if (!SetSecurityDescriptorSacl(psdAbsFileSD,
						bSaclPresent,       // fSaclPresent flag
						paclSacl,
						bSaclDefaulted)) {

		return FALSE;
	} 

	//
	// Set Owner into absolute format File SD
	//
	bOwnerDefaulted = (SE_OWNER_DEFAULTED & *psdcCtrl);
	if (!GetSecurityDescriptorOwner(psdSrelFileSD,
					&psidSidOwn,
					&bOwnerDefaulted)) {

		return FALSE;
	}
	if (!SetSecurityDescriptorOwner(psdAbsFileSD,
					psidSidOwn,
					bOwnerDefaulted)) {

		return FALSE;
	}

	//
	// Set Group into absolute format File SD
	//
	bGroupDefaulted = (SE_GROUP_DEFAULTED & *psdcCtrl);
	if (!GetSecurityDescriptorGroup(psdSrelFileSD,
					&psidSidPG,
					&bGroupDefaulted)) {

		return FALSE;
	}
	if (!SetSecurityDescriptorGroup(psdAbsFileSD,
					psidSidPG,
					bGroupDefaulted)) {

		return FALSE;
	}

	//
	// Now we can see if the old computer SID is embedded
	// in the security descriptor, and change it if necessary
	//
	CheckFileOwnershipForSID(psdAbsFileSD,lpszFullName);
	
	//
	// Check to see if SID is embedded in the DACL and then
	// SACL
	//
	CheckFileACLForSID( TRUE, psdAbsFileSD,lpszFullName);
	CheckFileACLForSID( FALSE, psdAbsFileSD, lpszFullName );
	return(TRUE);
}


//----------------------------------------------------------------------
//
// UpdateFileSID
//
// Updates the SIDs for the specified file or directory and recurses
// on subdirectories and files.
//
//----------------------------------------------------------------------
void UpdateFileSID( HWND hDlg, PTCHAR PathName )
{
	TCHAR			subName[MAX_PATH], searchName[MAX_PATH];
	HANDLE			dirHandle;
	WIN32_FIND_DATA foundFile;

	//
	// Process this file or directory
	//
	UpdateProgressDialog( hDlg, IDC_PATHNAME, PathName );
	CheckFileSID( PathName );

	//
	// Scan the files and/or directories if this is a directory
	//
	wsprintf( searchName, L"%s\\*.*", PathName );
	if( (dirHandle = FindFirstFile( searchName, &foundFile )) == 
		INVALID_HANDLE_VALUE  ) {

		//
		// Nothing to process
		//
		return;
	}

	do {

		if( wcscmp( foundFile.cFileName, L"." ) &&
			wcscmp( foundFile.cFileName, L".." )) {

			//
			// Process the subfile or directory
			//
			wsprintf( subName, L"%s\\%s", PathName, foundFile.cFileName );
			UpdateFileSID( hDlg, subName );
		}
	} while( FindNextFile( dirHandle, &foundFile ));

	FindClose( dirHandle );
}


//----------------------------------------------------------------------
//
// UpdateFileSystemSID
//
// Scans all the non-removable NTFS drives looking for SIDs
// that need to be changed in the security descriptors.
//
//----------------------------------------------------------------------
void UpdateFileSystemSID( void )
{
	DWORD	drives, curDrive;
	TCHAR	drivePath[16], startPath[16];
	TCHAR	fileSystem[256];
	HWND	fileDialog;

	//
	// Create a dialog to show the file traversal
	//
	fileDialog = CreateDialog( hInst, L"FILE", NULL,
							(DLGPROC) UpdateDlgProc );

	drives = GetLogicalDrives();
	curDrive = 2;  // start with 'C:'
	while( curDrive < 26 ) { 

		wsprintf( drivePath, L"%c:\\", curDrive + L'A');

		GetVolumeInformation( drivePath, NULL, 0, NULL, NULL,
								NULL, fileSystem, 256 );

		//
		// Only process local NTFS drives
		//
		if( ((1 << curDrive) & drives ) && 
			(GetDriveType( drivePath ) == DRIVE_FIXED) &&
			!wcsicmp( fileSystem, L"NTFS")) {

			wsprintf( startPath, L"%c:", curDrive + L'A');
			UpdateFileSID( fileDialog, startPath ); 	
		}
		curDrive++;
	}

	//
	// Delete the file dialog
	//
	DestroyWindow( fileDialog );
}


//===================================================================
//
// SAM AND SECURITY SID CHANGING ROUTINES
//
//===================================================================

//----------------------------------------------------------------------
//
// CopyKey
//
// Copies a subtree - only deals with the first level of keys and values
// since this is all that is needed for the keys that have
// the computer SID embedded in them.
//
//----------------------------------------------------------------------
VOID CopyKey( HKEY Root, TCHAR *Source, TCHAR *Destination, PSECURITY_DESCRIPTOR newSecDesc )
{
	HKEY					srcKey, dstKey, srcSubKey, dstSubKey;
	DWORD					status;
	PSECURITY_DESCRIPTOR	origDesc, subDesc;
	TCHAR					*Tail;
	FILETIME				Time;
	int						idx;
	TCHAR					origSourceName[1024];
	TCHAR					valName[1024];
	DWORD					valNameLen, valDataLen, valType, valDataBufsize = 1024;
	BYTE				*	valData = malloc( valDataBufsize );
	DWORD					nb;

	//
	// Open the source
	//
	wcscpy( origSourceName, Source );
	status = RegOpenKey( Root, Source, &srcKey );
	if( status != ERROR_SUCCESS ) {
		return;
	}

	//
	// Get the source security and set the new descriptor
	//
	GetRegAccess( srcKey, &origDesc );
	SetRegAccess( Root, Source, newSecDesc, &srcKey );

	//
	// Create a copy
	//
	status = RegCreateKey( Root, Destination, &dstKey );
	if( status != ERROR_SUCCESS ) {
		return;
	}

	//
	// Enumerate source values and create copies in the destination
	//
	wcscpy( valName, L"");
	idx = 0;
	valNameLen = 1024;
	valDataLen = valDataBufsize;
	status = RegQueryValue( srcKey, NULL, (PTCHAR) valData, &valDataLen );
	while ( status == ERROR_MORE_DATA )  {
		valDataBufsize = valDataLen += 1024;
		valData = realloc( valData, valDataBufsize );
		status = RegQueryValue( srcKey, NULL, (PTCHAR) valData, &valDataLen );
	}
	if ( status != ERROR_SUCCESS ) valDataLen = 0;
	do {

		//
		// Copy to the destination 
		//
		if( valDataLen ) 
			(void) RegSetValueEx( dstKey, valName, 0, valType, valData, valDataLen );

		valNameLen = 1024;
		valDataLen = valDataBufsize;
		status = RegEnumValue( srcKey, idx++, valName, &valNameLen,
					NULL, &valType,	valData, &valDataLen  );
		while ( status == ERROR_MORE_DATA )  {
			valDataBufsize = valDataLen += 1024;
			valData = realloc( valData, valDataBufsize );
			status = RegEnumValue( srcKey, idx-1, valName, &valNameLen,
									NULL, &valType,	valData, &valDataLen  );
		}

	} while ( status == ERROR_SUCCESS );

	//
	// Enumerate source subkeys and create copies in the destination
	//	
	Tail = Source + wcslen(Source);
	*Tail++ = '\\';

	idx = 0;
	while(1)  {

		nb = 1024;
		status = RegEnumKeyEx( srcKey, idx, Tail, &nb, NULL, NULL, NULL, &Time );
		if ( status == ERROR_NO_MORE_ITEMS )
			break;
		if ( status != ERROR_SUCCESS )
			break;

		//
		// Open the source subkey to get its security setting
		//
		status = RegOpenKey( srcKey, Tail, &srcSubKey );
		if( status != ERROR_SUCCESS ) {
			idx++;
			continue;
		}
		GetRegAccess( srcSubKey, &subDesc ); 
		SetRegAccess( srcKey, Tail, newSecDesc, &srcSubKey );
		RegDeleteKey( srcSubKey, NULL );
		RegCloseKey( srcSubKey );

		//
		// Create a copy of the key
		//
		RegCreateKey( dstKey, Tail, &dstSubKey );
		SetRegAccess( NULL, NULL, subDesc, &dstSubKey );
		RegCloseKey( dstSubKey );
		free( subDesc );

		//
		// Restart the emumeration
		//
		RegCloseKey( srcKey );
		status = RegOpenKey( Root, origSourceName, &srcKey );
		if( status != ERROR_SUCCESS ) {
			break;
		}		 
		idx = 0;
	}	

	//
	// Set copy of access on destination(which closes the keys)
	//
	SetRegAccess( NULL, NULL, newSecDesc, &dstKey );
	RegDeleteKey( srcKey, NULL );
	RegCloseKey( srcKey );
	RegCloseKey( dstKey );
	free( origDesc );
}


//----------------------------------------------------------------------
//
// ChangeComputerSID
//
// Recursively dive into the key, changing all occurrences of oldSubAuth
// to newSubAuth, whether its in a value, or in a key name.
//
//----------------------------------------------------------------------
BOOL ChangeComputerSID( HKEY Root, TCHAR *path, PSECURITY_DESCRIPTOR newSecDesc )
{
	LONG		status;
	HKEY		hKey;
	int         idx;
	TCHAR       Name[ 1024 ];
	TCHAR       *Tail, *newSubPtr;
	FILETIME	Time;
	TCHAR		newSubName[1024], textNewSubAuth[256];
	TCHAR		textSubAuth[256], textSubAuthTmp[256];
	TCHAR		valName[1024];
	DWORD		valNameLen, valDataLen, valType, valDataBufsize = 1024;
	BYTE	*	valData = malloc( valDataBufsize );
	DWORD		nb, j;	
	BOOLEAN		valChanged;
	PSECURITY_DESCRIPTOR oldSecDesc;

	//
	// Open key 
	//
	status = RegOpenKey( Root, path, &hKey );
	if ( status != ERROR_SUCCESS )
		return FALSE;

	//
	// Get the key's security descriptor and set the new one
	//
	if( !GetRegAccess( hKey, &oldSecDesc )) 
		return FALSE;
	if( !SetRegAccess( Root, path, newSecDesc, &hKey ) ) 
		return FALSE;

	//
	// Scan the values looking for the old SID and change it
	// to the new SID.
	//
	wcscpy( valName, L"");
	idx = 0;
	valNameLen = 1024;
	valDataLen = valDataBufsize;
	status = RegQueryValue( hKey, path, (PTCHAR) valData, &valDataLen );
	while ( status == ERROR_MORE_DATA )  {
		valDataBufsize = valDataLen += 1024;
		valData = realloc( valData, valDataBufsize );
		status = RegQueryValue( hKey, path, (PTCHAR) valData, &valDataLen );
	}
	do {

		//
		// If its large enough, scan for the old SID
		//
		valChanged = FALSE;
		if( valDataLen > SubAuthLength ) {

			for( j = 0; j < valDataLen - SubAuthLength + 1; j++ ) {

				if( !memcmp( valData + j, OldSubAuth, SubAuthLength )) {

					memcpy( valData + j, NewSubAuth, SubAuthLength );
					valChanged = TRUE;
				}
			}

			//
			// If we performed a replacement, update the value
			//
			if( valChanged ) {
				if( RegSetValueEx( hKey, valName, 0, valType, valData, valDataLen )
						!= ERROR_SUCCESS ) 
					return FALSE;
			}
		}

		valNameLen = 1024;
		valDataLen = valDataBufsize;
		status = RegEnumValue( hKey, idx++, valName, &valNameLen,
								NULL, &valType,	valData, &valDataLen  );
		while ( status == ERROR_MORE_DATA )  {
			valDataBufsize = valDataLen += 1024;
			valData = realloc( valData, valDataBufsize );
			status = RegEnumValue( hKey, idx-1, valName, &valNameLen,
									NULL, &valType,	valData, &valDataLen  );
		}
	
	} while ( status == ERROR_SUCCESS );


	//
	// Now we enumerate subkeys, because if the tail has 
	// the SID embedded in it, we have to copy the subtree
	//
	wcscpy( Name, path );
	Tail = Name + wcslen(Name);
	*Tail++ = L'\\';

	wsprintf( textSubAuth, L"" );
	for( j = 0; j < SubAuthLength / sizeof(DWORD); j++ ) {
		wsprintf( textSubAuthTmp, L"-%u", ((PDWORD) OldSubAuth)[j] );
		wcscat( textSubAuth, textSubAuthTmp );
	}

	wsprintf( textNewSubAuth, L"" );
	for( j = 0; j < SubAuthLength / sizeof(DWORD); j++ ) {
		wsprintf( textSubAuthTmp, L"-%u", ((PDWORD) NewSubAuth)[j] );
		wcscat( textNewSubAuth, textSubAuthTmp );
	}

	//
	// Get subkey names
	//
	for ( idx = 0;; idx ++ )  {

		nb = sizeof Name;
		status = RegEnumKeyEx( hKey, idx, Tail, &nb, NULL, NULL, NULL, &Time );
		if ( status == ERROR_NO_MORE_ITEMS )
			break;
		if ( status != ERROR_SUCCESS )
			break;		

		if( !ChangeComputerSID( Root, Name, newSecDesc ) ) {

			SetRegAccess( NULL, NULL, oldSecDesc, &hKey );
			RegCloseKey( hKey );
			free( oldSecDesc);
			return FALSE;
		}

		for( j = 0; j < wcslen( Tail ); j++ ) {
			if( !wcsncmp( textSubAuth, Tail+j, wcslen( textSubAuth ) ) ) {

				//
				// Got a match - copy the tree to the new name
				//
				wsprintf( newSubName, L"%s", Name );
				newSubPtr = (TCHAR *) ((DWORD) newSubName + (DWORD) Tail - (DWORD) Name + j*2); 
				wcscpy( newSubPtr, textNewSubAuth );
				wcscat( newSubPtr, Tail + j + wcslen( textSubAuth ));

				CopyKey( Root, Name, newSubName, newSecDesc );

				//
				// Restart the scan of the key
				//
				idx = -1;
				break;
			}
		}
	}
	
	//
	// Change scurity back and close key
	//
	SetRegAccess( NULL, NULL, oldSecDesc, &hKey );
	RegCloseKey( hKey );
	free( oldSecDesc);
	free( valData );
	return TRUE;
}


//===================================================================
//
// MAIN DIALOG AND WINMAN
//
//===================================================================

//----------------------------------------------------------------------
//
// WelcomeDialog
//
// Splash copyright, and choice.
//
//----------------------------------------------------------------------
LONG CALLBACK WelcomeDialog( HWND hDlg, UINT message, UINT wParam, LONG lParam )
{
	switch (message) {
		case WM_COMMAND:
			switch (wParam) {
				case IDOK:
					EndDialog( hDlg, DIALOG_OK ) ;
					return TRUE ;
				case IDCANCEL:
					if ( MessageBox( hDlg, L"Do you wish to abort the SID change?",
									L"NTSID", MB_ICONQUESTION|MB_YESNO )
									!= IDYES )
						break;
					EndDialog( hDlg, DIALOG_CANCEL );
					return TRUE;
				case IDC_ABOUT:
					DialogBox( hInst, L"ABOUT", NULL, (DLGPROC)AboutDlgProc );
					return TRUE;
			}
			break; 

		case WM_CLOSE:	
			EndDialog( hDlg, DIALOG_CANCEL );
			return TRUE;
	}

	return DefWindowProc( hDlg, message, wParam, lParam );
}

//----------------------------------------------------------------------
//  
// ShutDown
//
// Reboot the machine
//
//----------------------------------------------------------------------
BOOL ShutDown()
{
	HANDLE				hToken;    
	TOKEN_PRIVILEGES	tkp;      
	BOOL				fResult;     
 
	// 
	// Get the current process token handle 
	// so we can get shutdown privilege. 
	// 
	if (!OpenProcessToken(GetCurrentProcess(), 
	        TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken)) 
	    return FALSE;
 
	// Get the LUID for shutdown privilege. 
	LookupPrivilegeValue(NULL, SE_SHUTDOWN_NAME, 
	        &tkp.Privileges[0].Luid); 
 
	tkp.PrivilegeCount = 1;  
	tkp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED; 
	// Get shutdown privilege for this process. 
 	AdjustTokenPrivileges(hToken, FALSE, &tkp, 0, 
	    (PTOKEN_PRIVILEGES) NULL, 0); 

  	// Cannot test the return value of AdjustTokenPrivileges. 
	if (GetLastError() != ERROR_SUCCESS) 
	    return FALSE; 
 
	// Actually shutdown
	fResult = InitiateSystemShutdown(  NULL, NULL, 0, TRUE, TRUE );
 
	if (!fResult)  
	    return FALSE;
 
	// Disable shutdown privilege.
	tkp.Privileges[0].Attributes = 0; 
	AdjustTokenPrivileges(hToken, FALSE, &tkp, 0, 
	        (PTOKEN_PRIVILEGES) NULL, 0); 
 
	if (GetLastError() != ERROR_SUCCESS) 
	    return FALSE;

	return TRUE;
}


//----------------------------------------------------------------------
//
// DoSIDChange
//
// Actually directs the change of the computer's sid
//
//----------------------------------------------------------------------
void DoSIDChange( void )
{
	DWORD					Step = 0;
	DWORD					nb;
	DWORD					Status;
	HKEY					hKey;
	WNDCLASSEX				wndclass ;
	DWORD					valType;
	PBYTE					vData;
	HWND					generateDialog;
	PSECURITY_DESCRIPTOR	newSecDesc, oldSecDesc;

	//
	// create a generic window class
	//
	wndclass.cbSize			= sizeof( WNDCLASSEX );
	wndclass.style          = CS_HREDRAW | CS_VREDRAW ;
	wndclass.cbClsExtra     = 0 ;
	wndclass.cbWndExtra     = DLGWINDOWEXTRA ;
	wndclass.hInstance      = hInst;
	wndclass.hIcon          = LoadIcon (hInst, L"APPICON") ;
	wndclass.hIconSm		= LoadIcon (hInst, L"APPICON");
	wndclass.hCursor        = LoadCursor (NULL, IDC_ARROW) ;
	wndclass.hbrBackground  = (HBRUSH) (COLOR_BTNFACE+1);
	wndclass.lpszMenuName   = NULL ;

	wndclass.lpfnWndProc    = (WNDPROC) WelcomeDialog ;
	wndclass.lpszClassName  = L"Welcome";
	RegisterClassEx( &wndclass );
	Status = DialogBox( hInst, L"Welcome", NULL, 
					(DLGPROC)WelcomeDialog );

	if( Status == DIALOG_CANCEL ) return;

	//
	// Now, get the descriptor of HKLM and apply this to SECURITY
	//
	newSecDesc = GetRegSecDesc( HKEY_LOCAL_MACHINE, L"SYSTEM\\CurrentControlSet",
					DACL_SECURITY_INFORMATION );

	//
	// Read the last subauthority of the current computer SID
	//
	if( RegOpenKey( HKEY_LOCAL_MACHINE, L"SECURITY\\SAM\\Domains\\Account", 
		&hKey) != ERROR_SUCCESS ) {

		MessageBox( NULL, L"NTSID was unable to access the\nComputer's SID.",
									L"NTSID", MB_ICONERROR|MB_OK );	
		return;
	}
	GetRegAccess( hKey, &oldSecDesc );
	SetRegAccess( HKEY_LOCAL_MACHINE, L"SECURITY\\SAM\\Domains\\Account", 
				newSecDesc, &hKey );

	nb = 0;
	vData = NULL;
	RegQueryValueEx( hKey, L"V", NULL, &valType, vData, &nb );
	vData = (PBYTE) malloc( nb );
	Status = RegQueryValueEx( hKey, L"V", NULL, &valType, vData, &nb );
	if( Status != ERROR_SUCCESS ) {
		
		SetRegAccess( HKEY_LOCAL_MACHINE, L"SECURITY\\SAM\\Domains\\Account", 
				oldSecDesc, &hKey );
		MessageBox( NULL, L"NTSID was unable to access the\nComputer's SID.",
									L"NTSID", MB_ICONERROR|MB_OK );	
		return;
	}
	SetRegAccess( NULL, NULL, oldSecDesc, &hKey );
	RegCloseKey( hKey );

	//
	// Make sure that we're dealing with a SID we understand
	//
	if( *(PDWORD) (vData + nb - 4 * sizeof(DWORD)) != 0x00000015 ||
		*(PDWORD) (vData + nb - 5 *	sizeof(DWORD)) != 0x05000000 ) {
		MessageBox( NULL, L"NTSID does not undetstand the format of\nthe Computer's SID.",
									L"NTSID", MB_ICONERROR|MB_OK );	
		return;
	}
	memcpy( OldSubAuth, vData + nb - 3 * sizeof(DWORD), 3 * sizeof(DWORD));
	SubAuthLength = 3 * sizeof(DWORD);

	//
	// Generating a truly random SID takes awhile, so put up a dialog
	//
	generateDialog = CreateDialog( hInst, L"GENERATE", NULL,
							(DLGPROC) UpdateDlgProc );

	//
	// Generate new components for the 4 subauthorities
	//
	((PDWORD)NewSubAuth)[0]	= Random();
	((PDWORD)NewSubAuth)[1]	= Random();
	((PDWORD)NewSubAuth)[2]	= Random();

	DestroyWindow( generateDialog );

	// 
	// Replace the original computer SID with the one we generated
	//
	if( ChangeComputerSID( HKEY_LOCAL_MACHINE, L"SECURITY",
								newSecDesc )) {

		//
		// Update SIDs embedded in NTFS security descriptors
		//
		UpdateFileSystemSID();

		//
		// Update SIDs embedded in Registry keys
		//
		UpdateRegistrySID( newSecDesc );

		//
		// Change the profile information 
		//
		if( ChangeComputerSID( HKEY_LOCAL_MACHINE, 
					L"Software\\Microsoft\\Windows NT\\CurrentVersion\\ProfileList",
					newSecDesc ) ) {

			MessageBox( NULL, L"NTSID has successfully changed the\nComputer's SID.\n\nPress OK to reboot the computer.",
										L"NTSID", MB_ICONQUESTION|MB_OK );	

		} else {

			MessageBox( NULL, L"NTSID has successfully changed the\nComputer's SID but was unable\n"
							  L"to update profile information.\n\nPress OK to reboot the computer.",
										L"NTSID", MB_ICONQUESTION|MB_OK );	
		}

		//
		// After changing security descriptors, we have to reboot or users may run into 
		// situations where they can't access files they own!
		//
		ShutDown();

	} else {

		//
		// Something went wrong
		//
		MessageBox( NULL, L"NTSID was unable to change the\nComputer's SID.\n\nIt is suggested you restore your SAM and\nSECURITY keys from backup.",
									L"NTSID", MB_ICONERROR|MB_OK );	
	}
}


//----------------------------------------------------------------------
//  
// WinMain
//
//----------------------------------------------------------------------
int CALLBACK WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance,
        				LPSTR lpCmdLine, int nCmdShow )
{
	HANDLE				Client;
	TOKEN_PRIVILEGES  *	Priv;
	TCHAR				buf[ 1024 ];
	int					j;
	static TCHAR	*	P[] = {
		SE_TAKE_OWNERSHIP_NAME,
		SE_SECURITY_NAME,
		SE_BACKUP_NAME,
		SE_RESTORE_NAME
	};

	// 
	// Give ourselves privileges for manipulating security
	// 
	if ( ! OpenProcessToken( GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES, &Client ) ) {
		MessageBox( NULL, L"NTSID was unable to obtain necessary privilege to change the computer SID.",
								L"NTSID", MB_ICONERROR|MB_OK );
		return FALSE;
	}
	Priv = (void *)buf;
	Priv->PrivilegeCount = 0;
	for ( j = 0; j < sizeof P/sizeof P[0]; ++j )  {
		Priv->PrivilegeCount += 1;
		if ( ! LookupPrivilegeValue( NULL, P[j], &Priv->Privileges[j].Luid ) ) {
			MessageBox( NULL, L"NTSID was unable to obtain necessary privilege to change the computer SID.",
									L"NTSID", MB_ICONERROR|MB_OK );
			return FALSE;
		}
		Priv->Privileges[j].Attributes	= SE_PRIVILEGE_ENABLED;
	}
	if ( ! AdjustTokenPrivileges( Client, FALSE, Priv, 0, NULL, NULL ) ) {
		MessageBox( NULL, L"NTSID was unable to obtain necessary privilege to change the computer SID.",
								L"NTSID", MB_ICONERROR|MB_OK );
		return FALSE;
	}

	// 
	// do init stuff
	// 
	InitCommonControls();
	hInst = hInstance;

	// 
	// Ask the user if they *really* want to change the SID
	// 
	DoSIDChange();

	return 0;
}










